(*
 ________                        _____________
 __  ___/______ ____   _______   ___  __/___(_)_______ ___ _____
 _____ \ _  __ `/__ | / /_  _ \  __  /   __  / __  __ `__ \_  _ \
 ____/ / / /_/ / __ |/ / /  __/  _  /    _  /  _  / / / / //  __/
 /____/  \__,_/  _____/  \___/   /_/     /_/   /_/ /_/ /_/ \___/
 ___    _________________    ________              ______
 __ |  / /__  ____/___  /    ___  __/______ ______ ___  /________
 __ | / / _  /     __  /     __  /   _  __ \_  __ \__  / __  ___/
 __ |/ /  / /___   _  /___   _  /    / /_/ // /_/ /_  /  _(__  )
 _____/   \____/   /_____/   /_/     \____/ \____/ /_/   /____/

 ----------------------------------------------------------------
 Delphi components with source codes (c) Jakub Dusek
 ----------------------------------------------------------------
 TSTDrivesCombo & TSTDrivesList 1.11
 ----------------------------------------------------------------
 Two Delphi components that displays drives combo box or list box
 with associated icons. See included help file for detail
 description.
 ----------------------------------------------------------------
 This source code is not public domain and can be distributed
 only with this header. Mail me some bugs or good ideas about
 new functions.
 ----------------------------------------------------------------
 http://dusek.zde.cz                            j.dusek@seznam.cz
 ________________________________________________________________

*)

unit STDrives;

{$IFNDEF VER80}
 {$IFNDEF VER90}
  {$IFNDEF VER93}
    {$IFNDEF VER100}
      {$IFNDEF VER110}
        {$DEFINE ST4}
        {$IFNDEF VER120}
           {$IFNDEF VER125}
            {$DEFINE ST5}
          {$ENDIF}
        {$ENDIF}
      {$ENDIF}
    {$ENDIF}
  {$ENDIF}
 {$ENDIF}
{$ENDIF}

interface

uses
  Windows, StdCtrls, Messages, Controls, Classes, ShellApi,
  Graphics, SysUtils {$IFDEF ST4}, ImgList {$ENDIF};

type
  TMessageEvent = procedure (var Msg: TMsg; var Handled: Boolean) of object;

  TDriveType = (dtUnkNown, dtNonExisting, dtRemovable,
                dtFixed, dtRemote, dtCDROM, dtRamDisk);

  TDrivesType = set of TDriveType;

  TIconSize = (isSmall, isLarge);

  TTextCase = (tcLowerCase, tcUpperCase);

  TSTDrivesCombo = class (TCustomComboBox)
  private
    FAutoUpdate		: Boolean;
    FDisplayBrackets	: Boolean;
    FDisplayIcons	: Boolean;
    FDisplayLabel	: Boolean;
    FDisplaySysName	: Boolean;
    FDrive		: Char;
    FIconsSize		: TIconSize;
    FTextCase		: TTextCase;
    FDrivesType		: TDrivesType;

    procedure WndProc(var Msg: TMessage); override;

    procedure SetDisplayBrackets (Value: Boolean);
    procedure SetDisplayIcons (Value: Boolean);
    procedure SetDisplayLabel (Value: Boolean);
    procedure SetDisplaySysName (Value: Boolean);
    procedure SetDrive (Value: Char);
    procedure SetCase (Value: TTextCase);
    procedure SetIconsSize (Value: TIconSize);
    procedure SetDrivesType (Value: TDrivesType);
    procedure CMFontChanged (var Message: TMessage); Message CM_FONTCHANGED;
  protected
    procedure DropDown; override;
    procedure DrawItem (Index: Integer; Rect: TRect; State: TOwnerDrawState); override;
    procedure Change; override;

  public
    constructor Create (AOwner: TComponent); override;
    procedure CreateWnd; override;
    procedure Loaded; override;
  published
    property AutoUpdate		: Boolean read FAutoUpdate write FAutoUpdate default True;
    property DisplayBrackets	: Boolean read FDisplayBrackets write SetDisplayBrackets default True;
    property DisplayIcons	: Boolean read FDisplayIcons write SetDisplayIcons default True;
    property DisplayLabel	: Boolean read FDisplayLabel write SetDisplaylabel default True;
    property DisplaySysName	: Boolean read FDisplaySysName write SetDisplaySysName default True;
    property Drive		: Char read FDrive write SetDrive;
    property IconsSize		: TIconSize read FIconsSize write SetIconsSize;
    property TextCase		: TTextCase read FTextCase write SetCase default tcLowerCase;
    property DrivesType		: TDrivesType read FDrivesType write SetDrivesType default [dtRemovable,dtFixed, dtRemote, dtCDROM, dtRamDisk];
    procedure UpdateDrives;
    {$IFDEF ST4}
    property Anchors;
    property BiDiMode;
    property Constraints;
    property DragKind;
    property ParentBiDiMode;
    property OnEndDock;
    property OnStartDock;
    {$ENDIF}
    {$IFDEF ST5}
    property OnConTextPopup;
    {$ENDIF}
    {$IFNDEF VER90}
    property ImeMode;
    property ImeName;
    {$ENDIF}
    property Color;
    property Ctl3D;
    property DragCursor;
    property DragMode;
    property DropDownCount;
    property Enabled;
    property Font;
    property ParentColor;
    property ParentCtl3D;
    property ParentFont;
    property ParentShowHint;
    property PopupMenu;
    property ShowHint;
    property TabOrder;
    property Visible;
    property OnChange;
    property OnClick;
    property OnDblClick;
    property OnDragDrop;
    property OnDragOver;
    property OnDropDown;
    property OnEndDrag;
    property OnEnter;
    property OnExit;
    property OnKeyDown;
    property OnKeyPress;
    property OnKeyUp;
    property OnStartDrag;
  end;

  TSTDrivesList = class (TCustomListBox)
  private
    FDisplayBrackets	: Boolean;
    FDisplayIcons	: Boolean;
    FDisplayLabel	: Boolean;
    FDisplaySysName	: Boolean;
    FDrive		: Char;
    FIconsSize		: TIconSize;
    FTextCase		: TTextCase;
    FDrivesType		: TDrivesType;
    procedure SetDisplayBrackets (Value: Boolean);
    procedure SetDisplayIcons (Value: Boolean);
    procedure SetDisplayLabel (Value: Boolean);
    procedure SetDisplaySysName (Value: Boolean);
    procedure SetDrive (Value: Char);
    procedure SetCase (Value: TTextCase);
    procedure SetIconsSize (Value: TIconSize);
    procedure SetDrivesType (Value:TDrivesType);
    procedure CMFontChanged (var Message: TMessage); Message CM_FONTCHANGED;
  protected
    procedure DrawItem (Index: Integer; Rect: TRect; State: TOwnerDrawState); override;
    procedure Click; override;
  public
    constructor Create (AOwner: TComponent); override;
    procedure CreateWnd; override;
    procedure Loaded; override;
  published
    property DisplayBrackets	: Boolean read FDisplayBrackets write SetDisplayBrackets default True;
    property DisplayIcons	: Boolean read FDisplayIcons write SetDisplayIcons default True;
    property DisplayLabel	: Boolean read FDisplayLabel write SetDisplaylabel default True;
    property DisplaySysName	: Boolean read FDisplaySysName write SetDisplaySysName default True;
    property Drive		: Char read FDrive write SetDrive;
    property IconsSize		: TIconSize read FIconsSize write SetIconsSize;
    property TextCase		: TTextCase read FTextCase write SetCase default tcLowerCase;
    property DrivesType		: TDrivesType read FDrivesType write SetDrivesType default [dtRemovable,dtFixed, dtRemote, dtCDROM, dtRamDisk];
    procedure UpdateDrives;
    {$IFDEF ST4}
    property Anchors;
    property BiDiMode;
    property Constraints;
    property DragKind;
    property ParentBiDiMode;
    property OnEndDock;
    property OnStartDock;
    property OnMouseWheelDown;
    property OnMouseWheelUp;
    {$ENDIF}
    {$IFDEF ST5}
    property OnConTextPopup;
    {$ENDIF}
    {$IFNDEF VER90}
    property ImeMode;
    property ImeName;
    {$ENDIF}
    property Align;
    property BorderStyle;
    property Color;
    property Columns;
    property Ctl3D;
    property DragCursor;
    property DragMode;
    property Enabled;
    property ExtendedSelect;
    property Font;
    property IntegralHeight;
    property MultiSelect;
    property ParentColor;
    property ParentCtl3D;
    property ParentFont;
    property ParentShowHint;
    property PopupMenu;
    property ShowHint;
    property TabOrder;
    property Visible;
    property OnClick;
    property OnDblClick;
    property OnDragDrop;
    property OnDragOver;
    property OnEndDrag;
    property OnEnter;
    property OnExit;
    property OnKeyDown;
    property OnKeyPress;
    property OnKeyUp;
    property OnStartDrag;
  end;
  
  procedure Register;

var Larges, Smalls	: TImageList;
    SFI			: TSHFileInfo;
    SysIL		: uint;

implementation

procedure Register;
begin
  RegisterComponents ('ST Tools', [TSTDrivesCombo]);
  RegisterComponents ('ST Tools', [TSTDrivesList]);
end;

function GetDriveType (Drive: Char): TDriveType;
begin
  Result := TDriveType (Windows.GetDriveType (PChar (Drive + ':\')));
end;

function GetVolumeLabel (Drive: Char): string;
var VolLab		: array [0..MAX_PATH] of Char;
    MaxLength,SysFlag	: DWord;
    OldErrorMode	: Integer;
begin
  OldErrorMode := SetErrorMode (SEM_FAILCRITICALERRORS);
  try
    if GetVolumeInformation (PChar (Drive + ':\'), VolLab, 255,
                       nil, MaxLength, SysFlag, nil, 255) then Result := VolLab;
  finally
    SetErrorMode (OldErrorMode);
  end;
end;

procedure GetLogicalDrivesList (List: TStrings; DrivesType: TDrivesType);
 var
   Num		: Integer;
   Bits		: set of 0..25;
   Drive	: Char;
 begin
   List.Clear;
   Integer (Bits) := Windows.GetLogicalDrives;
   for Num := 0 to 25 do
   begin
     Drive := Char (Num + Ord ('A'));
     if (Num in Bits) and (GetDriveType (Drive) in DrivesType) then 
          List.Add (Drive);
   end;
end;

procedure DrawAssociatedImage (s: string; Cnvs: TCanvas; x, y: Integer; Size: TIconSize);
var Info	: TSHFileInfo;
    SSize	: Integer;
begin
  SSize := SizeOf (TSHFileInfo);
  FillChar (Info, SSize, 0);
  SHGetFileInfo (PChar (s), 0, Info, SSize, SHGFI_ICON or SHGFI_SMALLICON);
  case Size of
    isSmall: Smalls.Draw (Cnvs, x, y, Info.iIcon);
    isLarge: Larges.Draw (Cnvs, x, y, Info.iIcon);
  end;
end;

procedure DrawDriveItem(Drive: string; Canvas: TCanvas; Rect: TRect; DisplayLabel,
                        DisplaySysName, DisplayBrackets, DisplayIcons: Boolean;
                        TextCase: TTextCase; IconsSize: TIconSize);
var DriveLabel, DriveName	: string;
    x, y, Size			: Integer;
    Info			: TSHFileInfo;
begin
  DriveLabel := Drive + ': ';
  DriveName := Drive + ':\';
  if not DisplaySysName then
  begin
    if (DisplayLabel) and (GetDriveType (DriveLabel [1]) <> dtRemovable) then
    begin
      if DisplayBrackets then DriveLabel := DriveLabel + '[';
      DriveLabel := DriveLabel + GetVolumeLabel (DriveLabel [1]);
      if DisplayBrackets then DriveLabel := DriveLabel + ']';
    end;
    case TextCase of
      tcLowerCase: DriveLabel := LowerCase(DriveLabel);
      tcUpperCase: DriveLabel := UpperCase(DriveLabel);
    end;
  end
  else
  begin
    Size := SizeOf (TSHFileInfo);
    FillChar (Info, Size, 0);
    SHGetFileInfo (PChar (DriveName), 0, Info, Size, SHGFI_DISPLAYNAME);
    DriveLabel := Info.szDisplayName;
  end;
  Canvas.FillRect (Rect);
  Rect.Top := Rect.Top + 1;
  Rect.Bottom := Rect.Bottom - 1;
  Rect.Left := Rect.Left + 2;
  if DisplayIcons then
  begin
    case IconsSize of
      isSmall:
      begin
        y := Rect.Top + (((Rect.Bottom - Rect.Top) - 16) div 2);
        x := 20;
      end;
      isLarge:
      begin
        y := Rect.Top + (((Rect.Bottom-Rect.Top) - 32) div 2);
        x := 36;
      end;
    end;
    DrawAssociatedImage (DriveName, Canvas, Rect.Left, y, IconsSize);
  end
  else
    x := 2;
  Rect.Left := Rect.Left + x;
  DrawText (Canvas.Handle, PChar (DriveLabel), Length (DriveLabel), Rect,
            DT_SINGLELINE or DT_VCENTER or DT_NOPREFIX);
end;

function GetItemHeight (Font: TFont): Integer;
var
  DC		: HDC;
  Fnt		: HFont;
  Metrics	: TTextMetric;
begin
  DC := GetDC (0);
  try
    Fnt := SelectObject (DC, Font.Handle);
    GeTTextMetrics (DC, Metrics);
    SelectObject (DC, Fnt);
  finally
    ReleaseDC (0, DC);
  end;
  Result := Metrics.tmHeight + 1;
end;

constructor TSTDrivesCombo.Create (AOwner: TComponent);
begin
  inherited Create(AOwner);
  Style := csOwnerDrawFixed;
  FDrivesType := [dtRemovable,dtFixed, dtRemote, dtCDROM, dtRamDisk];
  AutoUpdate := True;
  DisplayIcons := True;
  DisplayBrackets := True;
  DisplayLabel := True;
  DisplaySysName := True;
end;

procedure TSTDrivesCombo.WndProc(var Msg: TMessage);
begin
  if (Msg.Msg = WM_DEVICECHANGE) then
    UpdateDrives;
  inherited;
end;

procedure TSTDrivesCombo.CreateWnd;
begin
  inherited CreateWnd;
  UpdateDrives;
end;

procedure TSTDrivesCombo.Loaded;
begin
  inherited Loaded;
  SetDrive (FDrive);
end;

procedure TSTDrivesCombo.DropDown;
begin
  if FAutoUpdate then UpdateDrives;
  inherited DropDown;
end;

procedure TSTDrivesCombo.SetIconsSize (Value: TIconSize);
begin
  FIconsSize := Value;
  case Value of
    isSmall:
    begin
      if GetItemHeight (Font) < 18 then ItemHeight := 18;
      Height := ItemHeight;
    end;
    isLarge:
    begin
      if GetItemHeight (Font) < 34 then ItemHeight := 34;
      Height := ItemHeight;
    end;
  end;
  RecreateWnd;
  Invalidate;
end;

procedure TSTDrivesCombo.SetDisplayIcons (Value: Boolean);
begin
  FDisplayIcons := Value;
  Invalidate;
end;

procedure TSTDrivesCombo.UpdateDrives;
var s		: string;
    Index	: Integer;
begin
  if ItemIndex <> - 1 then s := Items [ItemIndex];
  Items.Clear;
  GetLogicalDrivesList(Items, FDrivesType);
  if Items.Count <> 0 then
  begin
    Index := Items.IndexOf (s); 
    if Index <> - 1 then ItemIndex := Index
    else
      ItemIndex := 0;
    s := Items [ItemIndex];
    FDrive := s [1];
  end;
end;

procedure TSTDrivesCombo.Change;
var s: string;
begin
  s := Items [ItemIndex];
  FDrive := s [1];
  inherited Change;
end;

procedure TSTDrivesCombo.SetCase (Value: TTextCase);
begin
  FTextCase := Value;
  Invalidate;
end;

procedure TSTDrivesCombo.SetDrivesType (Value: TDrivesType);
begin
  FDrivesType := Value;
  UpdateDrives;
end;

procedure TSTDrivesCombo.SetDrive (Value: Char);
begin
  if Items.IndexOf (Value) <> - 1 then
  begin
    ItemIndex := Items.IndexOf (Value);
    FDrive := Value;
    if Assigned (OnChange) then OnChange (Self);
  end;
end;

procedure TSTDrivesCombo.SetDisplayBrackets (Value: Boolean);
begin
  FDisplayBrackets := Value;
  Invalidate;
end;

procedure TSTDrivesCombo.SetDisplayLabel (Value: Boolean);
begin
  FDisplayLabel := Value;
  Invalidate;
end;

procedure TSTDrivesCombo.SetDisplaySysName (Value: Boolean);
begin
  FDisplaySysName := Value;
  Invalidate;
end;

procedure TSTDrivesCombo.CMFontChanged (var Message: TMessage);
begin
  inherited;
  ItemHeight := GetItemHeight (Font);
  SetIconsSize (FIconsSize);
end;

procedure TSTDrivesCombo.DrawItem (Index: Integer; Rect: TRect; State: TOwnerDrawState);
begin
  DrawDriveItem (Items [Index], Canvas, Rect, FDisplayLabel,
                 FDisplaySysName, FDisplayBrackets,
                 FDisplayIcons, FTextCase, FIconsSize);
end;

constructor TSTDrivesList.Create (AOwner: TComponent);
begin
  inherited Create (AOwner);
  Style := lbOwnerDrawFixed;
  FDrivesType := [dtRemovable, dtFixed, dtRemote, dtCDROM, dtRamDisk];
  DisplayIcons := True;
  DisplayBrackets := True;
  DisplayLabel := True;
  DisplaySysName := True;
end;

procedure TSTDrivesList.CreateWnd;
begin
  inherited CreateWnd;
  UpdateDrives;
end;

procedure TSTDrivesList.Loaded;
begin
  inherited Loaded;
  SetDrive (FDrive);
end;

procedure TSTDrivesList.SetIconsSize (Value: TIconSize);
begin
  FIconsSize := Value;
  case Value of
    isSmall: if GetItemHeight (Font) < 18 then ItemHeight := 18;
    isLarge: if GetItemHeight (Font) < 34 then ItemHeight := 34;
  end;
  RecreateWnd;
  Invalidate;
end;

procedure TSTDrivesList.SetDisplayIcons (Value: Boolean);
begin
  FDisplayIcons := Value;
  Invalidate;
end;

procedure TSTDrivesList.UpdateDrives;
var s		: string;
    Index	: Integer;
begin
  if ItemIndex <> - 1 then s := Items [ItemIndex];
  Items.Clear;
  GetLogicalDrivesList( Items, FDrivesType);
  if Items.Count <> 0 then
  begin
    Index := Items.IndexOf (s);
    if Index <> - 1 then ItemIndex := Index
    else
      ItemIndex := 0;
    s := Items [ItemIndex];
    FDrive := s [1];
  end;
end;

procedure TSTDrivesList.Click;
var s: string;
begin
  s := Items [ItemIndex];
  FDrive := s [1];
  inherited Click;
end;

procedure TSTDrivesList.SetCase (Value: TTextCase);
begin
  FTextCase := Value;
  Invalidate;
end;

procedure TSTDrivesList.SetDrivesType (Value: TDrivesType);
begin
  FDrivesType := Value;
  UpdateDrives;
end;

procedure TSTDrivesList.SetDrive (Value: Char);
begin
  if Items.Indexof (Value) <> - 1 then
  begin
    ItemIndex := Items.IndexOf (Value);
    FDrive := Value;
    if Assigned (OnClick) then OnClick (Self);
  end;
end;

procedure TSTDrivesList.SetDisplayBrackets (Value: Boolean);
begin
  FDisplayBrackets := Value;
  Invalidate;
end;

procedure TSTDrivesList.SetDisplayLabel (Value: Boolean);
begin
  FDisplayLabel := Value;
  Invalidate;
end;

procedure TSTDrivesList.SetDisplaySysName (Value: Boolean);
begin
  FDisplaySysName := Value;
  Invalidate;
end;

procedure TSTDrivesList.CMFontChanged (var Message: TMessage);
begin
  inherited;
  ItemHeight := GetItemHeight (Font);
  SetIconsSize (FIconsSize);
end;

procedure TSTDrivesList.DrawItem (Index: Integer; Rect: TRect; State: TOwnerDrawState);
begin
  DrawDriveItem (Items [Index], Canvas, Rect, FDisplayLabel,
                 FDisplaySysName, FDisplayBrackets,
                 FDisplayIcons, FTextCase, FIconsSize);
end;

initialization
  Smalls := TImageList.Create (nil);
  SysIL := SHGetFileInfo ('', 0, SFI, SizeOf (SFI), SHGFI_SYSICONINDEX or SHGFI_SMALLICON);
  if SysIL <> 0 then
  begin
    Smalls.Handle := SysIL;
    Smalls.ShareImages := True;
    Smalls.DrawingStyle := dsTransParent;
  end;
  Larges := TImageList.Create (nil);
  SysIL := SHGetFileInfo ('', 0, SFI, SizeOf(SFI), SHGFI_SYSICONINDEX or SHGFI_LARGEICON);
  if SysIL <> 0 then 
  begin
    Larges.Handle := SysIL;
    Larges.ShareImages := True;
    Larges.DrawingStyle := dsTransParent;
  end;

finalization
  Larges.Free;
  Smalls.Free;
end.
